#include "all.h"

int editpos;
int editside;
int edithalfpos;
int editinsert;
int editmodified;
int virt_filesize;

extern int topview,botview,viewhexw,vieweof,filesize;
extern FILE *viewfile;
extern WINDOW *gui_msbar;
extern char *tmp_template;
extern char *viewfilename;
extern int currentdir,keep_msbar;

struct edit_region *st_region;  /* the first region */

void edit_setup(void) {
  editpos=topview;
  editside=1;
  edithalfpos=0;
  editinsert=0;
  editmodified=0;
  virt_filesize=view_physfilesize();

  st_region=malloc(sizeof(struct edit_region));
  if(!st_region) outofmem("edit_setup");
  st_region->len=view_physfilesize();
  st_region->text=NULL;
  st_region->phys=0;
  st_region->next=NULL;
  st_region->prev=NULL;
}

/* returns 1 to go back to view mode, 0 to stay in edit mode */
int edit_exitedit(void) {
  struct edit_region *rp,*nxt;
  if(editmodified) {
    switch(readln_getchar(gui_msbar,"Save changes? (y=yes n=no ^G=cancel)","yn\a")) {
    case 'y':
      if(edit_savefile()) return 0;
      break;
    case 'n': break;
    default:  set_cancel(); return 0;
    }
  }
  rp=st_region;
  while(rp) {
    nxt=rp->next;
    if(rp->text) free(rp->text);
    free(rp);
    rp=nxt;
  }
  return 1;
}

#define CHUNK_SIZ 1024
int edit_savefile(void) {
  char *tmpnam;
  char fnbuf[PATH_MAX+1];
  const char *outfile;
  int fout;
  int total,left,now,pos;
  char buf[CHUNK_SIZ];

  if(comm_prompt(fnbuf,PATH_MAX+1,viewfilename,1,"Save as: ",
     FM_TABCOMPLETE,PV_PATHNAME)) return 1;
  outfile=comm_parsepath(comm_makeabsolute(fnbuf,currentdir));
/* FIXME - don't need temp file unless we saved over original */
  tmpnam=strdup(tmp_template);
  if(!tmpnam) outofmem("edit_savefile");
  fout=mkstemp(tmpnam);
  if(-1==fout) {
    gui_mildbeep("Error: cannot open temporary file (%s)",strerror(errno));
    free(tmpnam);
    return 1;
  }
  left=total=virt_filesize;
  for(pos=0;left;) {
    gui_printmsg("writing: % 6dk, press ctrl-G to abort",pos/1024);
    keep_msbar=0;
    if(is_cancel()) {
      close(fout); unlink(tmpnam); free(tmpnam);
      return 1;
    }
    now=MIN(left,CHUNK_SIZ);
    if(now!=edit_readbuffer(pos,buf,now)) {
      gui_mildbeep("Error: cannot read from input");
      close(fout); unlink(tmpnam); free(tmpnam);
      return 1;
    }
    if(now!=write(fout,buf,now)) {
      gui_mildbeep("Error: cannot write to output (%s)",strerror(errno));
      close(fout); unlink(tmpnam); free(tmpnam);
      return 1;
    }
    pos+=now; left-=now;
  }
  close(fout);
  debuglog("edit_save moving:\n  <%s>\n  <%s>\n",tmpnam,outfile);
  fclose(viewfile);
  rename(tmpnam,outfile);
  free(tmpnam);
  viewfile=fopen(viewfilename,"r");
  if(viewfile==NULL) {
    crapout("edit_savefile","Error: cannot reopen file (%s)",strerror(errno));
  }
  fseek(viewfile,0,SEEK_END);
  filesize=ftell(viewfile);
  return 0;
}

void edit_flipside(void) {
  editside=!editside;
  edithalfpos=0;
}

void edit_keypress(char c) {
  struct edit_region *rp,*newreg,*nxt;
  int pos;
  int hexed;
  int ins=editinsert;

  editmodified=1;

  #define ED_MODIFY(v) do {    \
    if(editside==0) v=c;       \
    else if(edithalfpos==0)    \
      v=(v&0xf)+(hexed<<4);    \
    else                       \
      v=(v&0xf0)+hexed;        \
  } while(0)

  hexed=999; /* to keep gcc happy */
  if(editside==1) {
    if((c>='0')&&(c<='9')) hexed=c-'0';
    else
    if((c>='a')&&(c<='f')) hexed=c-'a'+10;
    else crapout("edit_keypress","value is not a hex digit");
    if(edithalfpos==1) ins=0;
  }

  edit_getregion(editpos,&pos,&rp);
  if(rp->text) {
    if(ins) {
      rp->text=realloc(rp->text,rp->len+1);
      if(!rp->text) outofmem("edit_keypress");
      rp->len++;
      memmove(rp->text+pos+1,rp->text+pos,rp->len-pos);
      rp->text[pos]=0;
      ED_MODIFY(rp->text[pos]);
      virt_filesize++;
    } else {
      ED_MODIFY(rp->text[pos]);
    }
  } else {
    edit_splitregion(pos,rp);
    newreg=malloc(sizeof(struct edit_region));
    if(!newreg) outofmem("edit_keypress");
    nxt=rp->next;
    newreg->next=nxt;
    newreg->prev=rp;
    nxt->prev=newreg;
    rp->next=newreg;
    newreg->len=1;
    newreg->text=malloc(sizeof(char));
    if(!newreg->text) outofmem("edit_keypress");
    if(ins) {
      newreg->text[0]=0;
      ED_MODIFY(newreg->text[0]);
      virt_filesize++;
    } else {
      if(editside==1) {
        fseek(viewfile,nxt->phys,SEEK_SET);
        if(1!=fread(newreg->text,1,1,viewfile))
          crapout("edit_keypress","fread failed");
      }
      nxt->len--;
      nxt->phys++;
      ED_MODIFY(newreg->text[0]);
    }
  }
  edit_right();
}

void edit_deletechar(void) {
  struct edit_region *rp;
  int pos;

  if(editpos==view_filesize()) return;
  editmodified=1;

  edit_getregion(editpos,&pos,&rp);
  if(rp->text) {
    rp->len--;
    memmove(rp->text+pos+1,rp->text,rp->len-pos);
    /* rp->text shrunk, but no need to realloc */
  } else {
    if(pos==0) {
      rp->phys++;
      rp->len--;
    } else if(pos==rp->len-1) {
      rp->len--;
    } else {
      edit_splitregion(pos,rp);
      rp->next->phys++;
      rp->next->len--;
    }
  }
  virt_filesize--;
  if(editpos==view_filesize()) editpos--;
}

void edit_backspace(void) {
  if(!editpos) return;
  editpos--;
  if(editpos<topview) view_scrollup();
  edit_deletechar();
}

void edit_splitregion(int pos,struct edit_region *rp) {
  struct edit_region *newreg;
  if(rp->text) crapout("edit_splitregion","cannot split a text region");

  debuglog("splitting region: pos=%d, len=%d\n",pos,rp->len);

  newreg=malloc(sizeof(struct edit_region));
  if(!newreg) outofmem("edit_splitregion");
  newreg->next=rp->next;
  newreg->prev=rp;
  rp->next=newreg;
  newreg->text=NULL;
  newreg->phys=rp->phys+pos;
  newreg->len=rp->len-pos;
  rp->len=pos;

  debuglog("    phys=%d, len=%d    phys=%d, len=%d\n",
           rp->phys,rp->len,newreg->phys,newreg->len);
}

void edit_getregion(int virt_pos,int *pos,struct edit_region **rp) {
  *pos=0;
  *rp=st_region;
  while((*pos+(*rp)->len)<=virt_pos) {
    *pos+=(*rp)->len;
    *rp=(*rp)->next;
    if(!*rp) crapout("edit_getregion","rp ran out");
  }
  *pos=virt_pos-*pos;
}

int edit_readbuffer(int virt_pos,char *buf,int size) {
  struct edit_region *rp;
  int pos;
  int numout;
  int numnow;

  debuglog("edit_readbuffer: virt_pos=%d, size=%d, filesize=%d\n",
           virt_pos,size,view_filesize());

  if(virt_pos+size>view_filesize()) size=view_filesize()-virt_pos;
  edit_getregion(virt_pos,&pos,&rp);

  debuglog("    pos=%d, size=%d\n",pos,size);

  numout=0;
  while(numout<size) {
    if(!rp) crapout("edit_readbuffer","rp ran out (2)");
    numnow=MIN(size-numout,rp->len-pos);
    if(rp->text) memcpy(buf+numout,rp->text+pos,numnow);
    else {
      fseek(viewfile,rp->phys+pos,SEEK_SET);
      if(numnow>fread(buf+numout,1,numnow,viewfile))
        crapout("edit_readbuffer","fread came up short");
    }
    numout+=numnow;
    rp=rp->next;
    pos=0;
  }
  debuglog("    numout=%d, size=%d\n",numout,size);
  if(numout!=size) crapout("edit_readbuffer","numout!=size");
  return numout;
}

void edit_insert(void) {
  editinsert=!editinsert;
}

void edit_left(void) {
  if(!editpos && editside && !edithalfpos) return;
  if(editside) {
    if(edithalfpos) {edithalfpos=0;}
    else            {edithalfpos=1; editpos--;}
  } else editpos--;
  if(editpos<0) editpos=0;
  if(editpos<topview) view_scrollup();
}

void edit_right(void) {
  if((editpos==view_filesize()-1)&&edithalfpos) return;
  if(editside) {
    if(edithalfpos) {edithalfpos=0; editpos++;}
    else            {edithalfpos=1;}
  } else editpos++;
  if(editpos>botview) view_scrolldown();
}

void edit_up(void) {
  if(editpos-viewhexw<0) return;
  editpos-=viewhexw;
  if(editpos<0) {editpos=0; edithalfpos=0;}
  if(editpos<topview) view_scrollup();
}

void edit_down(void) {
  editpos+=viewhexw;
  if(editpos>botview) view_scrolldown();
  if(editpos>view_filesize()-1) {editpos=view_filesize()-1; edithalfpos=1;}
}

void edit_pageup(void) {
  int delta;
  delta=editpos-topview;
  view_pageup();
  editpos=delta+topview;
  if(editpos<0) {editpos=0; edithalfpos=0;}
}

void edit_pagedn(void) {
  int delta;
  delta=editpos-topview;
  view_pagedn();
  editpos=delta+topview;
  if(editpos>view_filesize()-1) {editpos=view_filesize()-1; edithalfpos=1;}
}

void edit_home(void) {
  view_home();
  editpos=0;
  edithalfpos=0;
}

void edit_end(void) {
  view_end();
  editpos=view_filesize()-1;
  edithalfpos=1;
}
